"""
Transition-specific calculations

References:
  [1] agrup.plez7.f
"""

from collections import OrderedDict


__all__ = ["calc_qgbd_tio_like"]


def calc_qgbd_tio_like(molconsts, v_lo):
    """
    Calculates qv, gv, bv, dv in TiO-like fashion

    Based on Fortran source 'agrup.plez7.f'

    Args:
        molconsts: dict-like object with keys:
            "omega_e"
            "omega_ex_e"
            "omega_ey_e"
            "B_e"
            "alpha_e"
            "D_e"
            "beta_e"

        v_lo: (integer) low transition level

    Returns: qbdg, i.e., a dictionary with keys:
            "qv", "gv", "bv", "dv", "gzero"
    """

    omega_e = molconsts["state2l_omega_e"]
    omega_ex_e = molconsts["state2l_omega_ex_e"]
    omega_ey_e = molconsts["state2l_omega_ey_e"]
    B_e = molconsts["state2l_B_e"]
    D_e = molconsts["state2l_D_e"]
    alpha_e = molconsts["state2l_alpha_e"]
    beta_e = molconsts["state2l_beta_e"]

    if int(v_lo) != v_lo:
        raise ValueError("Argument 'v_lo' must be an integer")

    gzero = omega_e / 2.0 - omega_ex_e / 4.0 + omega_ey_e / 8.0
    v_lo5 = v_lo + 0.5

    # Franck-Condon factor. Currently always incorporated into "SJ"
    qv = 1.
    # rotational constant "B_v"
    bv = B_e - alpha_e * v_lo5
    # rotational constant "D_v"
    dv = (D_e + beta_e * v_lo5) * 1.0e+06
    # rotational term " G_A"
    gv = omega_e * v_lo5 - omega_ex_e * v_lo5 ** 2 + omega_ey_e * v_lo5 ** 3 - gzero

    qgbd = OrderedDict((("qv", qv), ("gv", gv), ("bv", bv), ("dv", dv), ("gzero", gzero)))

    return qgbd



"""TurboSpectrum linelist format. Prepared for molecules only"""


__all__ = ["PlezAtomicLine", "PlezMolecularLine", "PlezSpecies", "FilePlezLinelist"]

import a99
from f311 import DataFile
import io
import os
from collections import defaultdict
import re


# Will update progress status every time the following number of lines is read from file
_PROGRESS_INDICATOR_PERIOD = 1030 * 5


class PlezLine(object):
    def __init__(self):
        self.lambda_ = 1.
        self.chiex = 1.
        self.loggf = 1.
        self.gu = 1.
        self.fdamp = 1.
        self.raddmp = 1.


class PlezAtomicLine(PlezLine):
    pass


class PlezMolecularLine(PlezLine):
    def __init__(self):
        PlezLine.__init__(self)
        self.branch = ""
        self.vl = 0
        self.v2l = 0
        self.Jl = 0.
        self.J2l = 0.


class PlezSpecies(object):
    def __init__(self, name="", lines=None):
        if lines is None:
            lines = []

        self.name = name
        self.lines = lines


@a99.froze_it
class FilePlezLinelist(DataFile):
    """
    Plez molecular lines file

    **Note** Plez file so far refers to one molecule only.

    **Note** Load only
    """

    def __init__(self):
        DataFile.__init__(self)

        self.molecules = defaultdict(lambda: PlezSpecies())
        self.atoms = defaultdict(lambda: PlezSpecies())

    def _do_load(self, filename):

         # '0107.000014          '  1       12150
         # 'NH A-X Fernando et al., 2018, JQSRT, 217,29 '
         #    3357.55219    0.004   -2.205 .00    3.  0.000E+00 'qR23(0)        0    1.0  e  F2e  -  0    0.0  e  F3e A            X           '
         #    3349.33946    0.004   -2.625 .00    3.  0.000E+00 'rR3(0)         0    1.0  e  F3e  -  0    0.0  e  F3e       A            X     '

        with open(filename, "r") as h:
            self._do_load_h(h, filename)

    def _do_load_h(self, h, filename):
        r = 0  # counts rows of file
        ii = 0

        expr_branch = re.compile("[PQRS]\d*")  # regular expression to find the branch

        # states
        EXP_SPECIES = 0  # expecting species definition
        EXP_NAME = 1  # expecting name of species
        EXP_ANY = 2

        state = EXP_SPECIES
        is_atom = None
        species = None
        flag_break = False

        try:
            while True:
                s = h.readline().strip()
                if len(s) == 0:
                    flag_break = True

                if not flag_break:
                    if s[0] == "'":
                        if state == EXP_SPECIES or state == EXP_ANY:
                            # New species to come

                            # Determine whether atom or molecule
                            atnumber = int(s[1:s.index(".")])
                            is_atom = atnumber <= 92

                            state = EXP_NAME

                        elif EXP_NAME:
                            name = s.strip("'")
                            species = self.atoms[name] if is_atom else self.molecules[name]

                            a99.get_python_logger().info(
                                 "Loading '{}': species '{}'".format(filename, name))

                            state = EXP_ANY

                    else:
                        if state != EXP_ANY:
                            raise RuntimeError("Not expecting line right now")
                        # It is a line

                        # Read 6 numeric values then something between quotes
                        line = PlezAtomicLine() if is_atom else PlezMolecularLine()

                        line.lambda_, line.chiex, line.loggf, line.gu, line.fdamp, line.raddmp = \
                            [float(x) for x in s[:s.index("'")].split()]

                        if not is_atom:
                            # Get the branch
                            tmp = s[s.index("'"):].strip("'").split()
                            _branch = tmp[0]
                            line.branch = expr_branch.search(_branch).group()
                            line.Jl = float(tmp[2])
                            line.vl = int(tmp[1])
                            line.v2l = int(tmp[6])
                            line.J2l = float(tmp[7])


                        species.lines.append(line)

                if flag_break:
                    break

                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    # a99.get_python_logger().info(
                    #      "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0

        except Exception as e:
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e





"""
HITRAN-specific conversion

References:
  [1] Rothman, Laurence S., et al. "The HITRAN 2004 molecular spectroscopic database."
      Journal of Quantitative Spectroscopy and Radiative Transfer 96.2 (2005): 139-204.
  [2] extraehitran.f (First version: Jorge Melendez-Moreno, January 1999)
"""

# TODO not working


from .convlog import *
from collections import OrderedDict
import airvacuumvald as avv
import pyfant


__all__ = ["hitran_to_sols"]


_OH_BRANCH_MAP_OH = {"QQ": "Q", "PP": "P", "RR": "R"}



def _parse_local_lower_quanta_OH(Q2l):
    """
    Parses the "Lower-state 'local' quanta" for 'OH' (see [1], Table 4, group 2, observation 'b')

    Args:
        Q2l: 15-character string containing the lower-state 'local' quanta information of molecular
        line. Example: " PP 12.5ff     "

    Returns: (Br, J2l), where
        Br: branch ("P"/"Q"/"R")
        J2l: inferior "quantum number associated with the total angular momentum excluding nuclear
             spin" [1]

    **Note** will raise if cannot determine Br (branch)
    """
    brinfo = Q2l[1:3]
    Br = _OH_BRANCH_MAP_OH.get(brinfo)
    if Br is None:
        raise ValueError("Cannot derive branch from '{}' ((local lower quanta) = '{}')".
                         format(brinfo, Q2l))
    J2l = float(Q2l[3:8])
    return Br, J2l

def _parse_local_lower_quanta_group2(Q2l):
    """
    Parses the "lower-state 'local' quanta" for molecules in [1], Table 4, group 2

    Args, Returns: see _parse_local_lower_quanta_OH()
    """

    # TODO test, not sure
    Br = Q2l[5:6]
    J2l = float(Q2l[6:9])
    return Br, J2l


def _parse_local_lower_quanta_group6(Q2l):
    """[...] [1], Table 4, group 6"""

    # TODO test, not sure
    Br = Q2l[3:4]
    J2l = float(Q2l[4:9])
    return Br, J2l


def _parse_global_quanta_class1(V):
    """
    Parses the [upper/lower] "global quanta" ([1], Table 3, class 1

    Args:
        V: 15-character string containing the 'global' quanta information of molecular
           line. Example: "       X3/2   8"

    Returns: v1: "quantum number associated with the first normal mode of vibration" [1]
    """
    return int(V[13:15])


def _parse_global_quanta_class2(V):
    """[...] [1], Table 3, class 2"""
    return int(V[13:15])

def _parse_global_quanta_class3(V):
    return int(V[13:15])


__GROUP_MAP = (
    (("OH",),
     _parse_local_lower_quanta_OH),
    (("CO2", "N2O", "CO", "HF", "HCl", "HBr", "HI", "OCS", "N2", "HCN", "C2H2", "NO+"),
     _parse_local_lower_quanta_group2),
    (("NO", "ClO"),
     _parse_local_lower_quanta_group6),
)
_GROUP_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __GROUP_MAP], []))

__CLASS_MAP = (
    (("CO", "HF", "HCl", "HBr", "HI", "N2", "NO+"), _parse_global_quanta_class1),
    (("O2",), _parse_global_quanta_class2),
    (("NO", "OH", "ClO"), _parse_global_quanta_class3),
)
_CLASS_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __CLASS_MAP], []))


def hitran_to_sols(molconsts, lines, qgbd_calculator):
    """
    Converts HITRAN molecular lines data to PFANT "sets of lines"

    Args:
        molconsts: a dict-like object combining field values from tables 'molecule', 'state',
                    'pfantmol', and 'system' from a FileMolDB database
        lines: item from rom hapi.LOCAL_TABLE_CACHE (a dictionary)
        qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
                         e.g., calc_qbdg_tio_like()

    Returns: (a list of ftpyfant.SetOfLines objects, a MolConversionLog object)
    """

    def append_error(msg):
        log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

    # C     BAND (v',v'')=(VL,V2L)
    # C     WN: vacuum wavenumber   WL : air wavelength
    # C     J2L: lower state angular momentum number
    # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
    # C     ramo: branch, for P: ID=1, for R: ID=2
    # C     name: file name/ coisovLv2L
    # C     HL: Honl-London factor
    # C     FR: oscillator strength

    header = lines["header"]
    data = lines["data"]

    S = molconsts.get_S2l()
    DELTAK = molconsts.get_deltak()
    formula = molconsts["formula"]

    n = header["number_of_rows"]
    log = MolConversionLog(n)
    sols = OrderedDict()  # one item per (vl, v2l) pair

    try:
        f_class = _CLASS_DICT[formula]
    except KeyError:
        log.errors.append("Cannot determine Table 3 'class' (Rothman et al. 2005) for formula '{}'".
                          format(formula))
        log.flag_ok = False

    try:
        f_group = _GROUP_DICT[formula]
    except KeyError:
        log.errors.append("Cannot determine Table 4 'group' (Rothman et al. 2005) for formula '{}'".
                          format(formula))
        log.flag_ok = False

    if not log.flag_ok:
        return [], log

    for i in range(n):

        try:
            nu = data["nu"][i]
            wl = avv.vacuum_to_air(1e8/nu)
            Br, J2l = f_group(data["local_lower_quanta"][i])
            # TODO assuming singlet!!!
            Jl = ph.singlet.quanta_to_branch(Br, J2l)
            V = f_class(data["global_upper_quanta"][i])
            V_ = f_class(data["global_lower_quanta"][i])
            A = data["a"][i]

            # A seguir normalizacion de factor de Honl-london (HLN)
            Normaliza = 1/((2.0*J2l+1)*(2.0*S+1)*(2.0-DELTAK))

            # A seguir teremos a forca de oscilador
            # mas q sera normalizada segundo o programa da Beatriz
            gf = Normaliza*1.499*(2*Jl+1)*A/(nu**2)

            J2l_pfant = int(J2l)  # ojo, estamos colocando J2L-0.5! TODO ask BLB: we write J2l or J2l-.5? (20171114 I think this is wrong, molecules.dat has decimal values)
        except Exception as e:
            log.errors.append("#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), a99.str_exc(e)))
            continue

        sol_key = (V, V_)  # (v', v'') transition (v_sup, v_inf)
        if sol_key not in sols:
            qgbd = qgbd_calculator(molconsts, sol_key[1])
            qqv = qgbd["qv"]
            ggv = qgbd["gv"]
            bbv = qgbd["bv"]
            ddv = qgbd["dv"]

            sols[sol_key] = pyfant.SetOfLines(V, V_, qqv, ggv, bbv, ddv, 1.)
        sol = sols[sol_key]
        sol.append_line(wl, gf, J2l_pfant, Br)

    return (list(sols.values()), log)



from .convlog import *
from .conv import *
import pyfant

__all__ = ["ConvPlez"]


class ConvPlez(Conv):
    """Converts Plez molecular lines data to PFANT "sets of lines"

        Args:
            flag_hlf: Whether to calculate the gf's using Honl-London factors or
                      use Plez's loggf instead

                      ***Note** old format (FilePlezMoleculeOld) does not have loggf information,
                                therefore the only way to work with the latter file type is
                                flag_hlf==True

            flag_normhlf: Whether to multiply calculated gf's by normalization factor

            flag_fcf: Whether to multiply calculated gf's by Franck-Condon Factor

            flag_quiet: Will not log exceptions when a molecular line fails

            name: name of molecule+system, e.g., "NH A-X PGopher"
    """

    def __init__(self, flag_hlf=False, flag_fcf=False, flag_quiet=False, flag_special_fcf=False,
                 fcfs=None, name="", *args, **kwargs):
        Conv.__init__(self, *args, **kwargs)
        self.flag_hlf = flag_hlf
        self.flag_fcf = flag_fcf
        self.flag_quiet = flag_quiet
        self.flag_special_fcf = flag_special_fcf
        self.fcfs = fcfs
        self.name = name

    def _make_sols(self, file):

        def append_error(msg):
            log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

        if not isinstance(file, pyfant.FilePlezLinelist):
            raise TypeError("Invalid type for argument 'fileobj': {}".format(type(file).__name__))

        lines = file.molecules[self.name].lines
        n = len(lines)

        if n == 0:
            raise RuntimeError("Species '{}' has zero lines".format(self.name))

        STATEL = self.molconsts["from_label"]
        STATE2L = self.molconsts["to_label"]

        mtools = self.kovacs_toolbox()

        # Prepares result
        sols = ConvSols(self.qgbd_calculator, self.molconsts)
        log = MolConversionLog(n)

        for i, line in enumerate(lines):
            branch = line.branch

            try:
                sj = 1.

                if self.flag_hlf:
                    try:
                        hlf = mtools.get_sj(line.vl, line.v2l, line.J2l, branch)
                    except pyfant.NoLineStrength:
                        log.skip_reasons["Cannot calculate HLF"] += 1
                        continue

                    if hlf < 0:
                        log.skip_reasons["Negative SJ"] += 1
                        continue

                    sj *= hlf

                else:
                    sj *= 10**line.loggf

                if self.flag_fcf:
                    sj *= self._get_fcf(line.vl, line.v2l, self.flag_special_fcf)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)
                continue

            sols.append_line(line, sj, branch)

        return sols, log



import pyfant
from .calc_qgbd import *
from .convlog import *
import datetime
from collections import OrderedDict
import math
import pyfant

__all__ = ["Conv", "ConvSols"]


_DEFAULT_QGBD_CALCULATOR = calc_qgbd_tio_like


class ConvSols(OrderedDict):
    """Stores (vl, v2l) as keys, SetOfLines as values"""

    def __init__(self, qgbd_calculator, molconsts):
        OrderedDict.__init__(self)
        self.qgbd_calculator = qgbd_calculator if qgbd_calculator else _DEFAULT_QGBD_CALCULATOR
        self.molconsts = molconsts

    def append_line(self, line, gf_pfant, branch):
        """Use to append line to object"""
        sol_key = (line.vl, line.v2l)

        if sol_key not in self:
            qgbd = self.qgbd_calculator(self.molconsts, line.v2l)
            self[sol_key] = pyfant.SetOfLines(line.vl, line.v2l,
                                          qgbd["qv"], qgbd["gv"], qgbd["bv"], qgbd["dv"], 1.)

        self[sol_key].append_line(line.lambda_, gf_pfant, line.J2l, branch)


class Conv(object):
    # TODO there is a good chance that flag_hlf, flag_fcf, etc. will be moved to this class
    """Molecular lines converter base class

    Args:
        qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
                         Default: calc_qbdg_tio_like

        molconsts: a dict-like object combining field values from tables 'molecule', 'state',
                    'pfantmol', and 'system' from a FileMolDB database

        fcfs: Franck-Condon Factors (dictionary of floats indexed by (vl, v2l))
    """

    def __init__(self, qgbd_calculator=None, molconsts=None, flag_normhlf=True, fcfs=None):
        self.qgbd_calculator = qgbd_calculator if qgbd_calculator else calc_qgbd_tio_like
        self.molconsts = molconsts
        self.fcfs = fcfs
        self.flag_normhlf = flag_normhlf

    def make_file_molecules(self, lines):
        """
        Builds a FileMolecules object

        Args:
            lines: in most cases, a DataFile instance, but may be anything, since the molecular
                   lines data extraction is performed by a particular descendant class.
                   For example, the HITRAN converter expects a structure from the HAPI

        Returns:
            f, log: FileMolecules, MolConversionLog instances
        """

        # Runs specific conversor to SetOfLines
        sols, log = self._make_sols(lines)

        assert isinstance(sols, ConvSols)
        assert isinstance(log, MolConversionLog)

        sols_list = list(sols.values())
        sols_list.sort(key= lambda sol: sol.vl*1000+sol.v2l)

        mol = pyfant.molconsts_to_molecule(self.molconsts)
        mol.sol = sols_list
        f = pyfant.FileMolecules()
        now = datetime.datetime.now()
        f.titm = "PFANT molecular lines file. Created by f311.convmol.Conv.make_file_molecules() @ {}".format(now.isoformat())
        f.molecules = [mol]
        return f, log

    def kovacs_toolbox(self):
        """Wraps f311.physics.multiplicity.kovacs_toolbox()"""
        return pyfant.kovacs_toolbox(self.molconsts, flag_normalize=self.flag_normhlf)

    # Must reimplement thig
    def _make_sols(self, lines):
        """Converts molecular lines into a list of SetOfLines object

        Args:
            lines: see make_file_molecules()

        Returns:
            sols, log: ConvSols, MolConversionLog
        """
        raise NotImplementedError()

    ######
    # Gear

    def _calculate_qgbd(self, v2l):
        return self.qgbd_calculator(self.molconsts, v2l)

    def _get_fcf(self, vl, v2l, flag_special=False):
        try:
            fcf = self.fcfs[(vl, v2l)]

            if flag_special:
                fcf = fcf * 10 ** -(math.floor(math.log10(fcf)) + 1)

        except KeyError as e:
            raise KeyError("FCF not available for (vl, v2l) = ({}, {})".format(vl, v2l))
        return fcf



from collections import Counter
import a99

@a99.froze_it
class MolConversionLog(object):

    @property
    def num_lines_skipped(self):
        """Sums all counts in self.skip_reasons"""
        return sum(self.skip_reasons.values())


    def __init__(self, num_lines=0, flag_ok=True):
        # Number of lines in input file
        self.num_lines = num_lines
        # Whether the conversion was considered successful (despite errors)
        self.flag_ok = flag_ok
        # List of error strings
        self.errors = []
        # Counter: {"reason": number_of_times, ...}
        # The key should be a descriptive reason, however shortish
        # Access like this:
        #     >>> log.skip_reasons["FCF not found for (vl,v2l)"] += 1
        #
        self.skip_reasons = Counter()




from .convlog import *
from .conv import *
import pyfant

__all__ = ["ConvKurucz"]


class ConvKurucz(Conv):
    """Converts Kurucz molecular lines data to PFANT "sets of lines"

        Args:
            flag_hlf: Whether to calculate the gf's using Honl-London factors or
                      use Kurucz's loggf instead

                      ***Note** old format (FileKuruczMoleculeOld) does not have loggf information,
                                therefore the only way to work with the latter file type is
                                flag_hlf==True

            flag_normhlf: Whether to multiply calculated gf's by normalization factor

            flag_fcf: Whether to multiply calculated gf's by Franck-Condon Factor

            flag_quiet: Will not log exceptions when a molecular line fails

            flag_spinl: Whether or not to use the spinl of the Kurucz line list
                        for branch determination (spin2l is always used). Effect:

                - True: possible branches are P1, P12, P21, P2 and repeat for Q, R
                        (12 possibilities in total)

                - False: possible branches are P1, P2 and repeat for Q, R
                         (6 possibilities in total)

            iso: (int or None) isotope. If specified as int, only that isotope will be filtered;
                 otherwise, all isotopes in file will be included. Isotope is
                 field KuruczMolLine.iso (see KuruczMolLine, FileKuruczMol).

                 **Note** old Kurucz file format does not have the isotope information, therefore,
                          iso must be None or the thing will not work

    """

    def __init__(self, flag_hlf=False, flag_fcf=False, flag_quiet=False, flag_special_fcf=False,
                 flag_spinl=False, fcfs=None, iso=None, *args, **kwargs):
        Conv.__init__(self, *args, **kwargs)
        self.flag_hlf = flag_hlf
        self.flag_fcf = flag_fcf
        self.flag_quiet = flag_quiet
        self.flag_spinl = flag_spinl
        self.flag_special_fcf = flag_special_fcf
        self.fcfs = fcfs
        self.iso = iso

    # TODO looks like this routine is kindda generic
    # TODO could dismember this later
    def _make_sols(self, lines):

        def append_error(msg):
            log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

        if not isinstance(lines, pyfant.FileKuruczMoleculeBase):
            raise TypeError("Invalid type for argument 'fileobj': {}".format(type(lines).__name__))
        assert self.flag_hlf or isinstance(lines, pyfant.FileKuruczMolecule), \
               "Old-format file does not contain loggf, must activate Hönl-London factors"

        lines = lines.lines
        n = len(lines)

        STATEL = self.molconsts["from_label"]
        STATE2L = self.molconsts["to_label"]

        mtools = self.kovacs_toolbox()

        # Prepares result
        sols = ConvSols(self.qgbd_calculator, self.molconsts)
        log = MolConversionLog(n)

        for i, line in enumerate(lines):
            if self.iso and line.iso != self.iso:
                log.skip_reasons["Isotope {}".format(line.iso)] += 1
                continue

            if line.statel != STATEL or line.state2l != STATE2L:
                log.skip_reasons["Transition {}-{}".format(line.statel, line.state2l)] += 1
                continue


            branch = mtools.quanta_to_branch(line.Jl, line.J2l,
                spinl=(None if not self.flag_spinl else line.spinl), spin2l=line.spin2l)


            # This was only a test, but filters like these may be useful if line.spin2l != line.spinl:
            #     log.skip_reasons["Different spin"] += 1
            #     continue


            try:
                sj = 1.

                if self.flag_hlf:
                    try:
                        hlf = mtools.get_sj(line.vl, line.v2l, line.J2l, branch)
                    except pyfant.NoLineStrength:
                        log.skip_reasons["Cannot calculate HLF"] += 1
                        continue

                    if hlf < 0:
                        log.skip_reasons["Negative SJ"] += 1
                        continue

                    sj *= hlf

                else:
                    sj *= 10**line.loggf

                if self.flag_fcf:
                    sj *= self._get_fcf(line.vl, line.v2l, self.flag_special_fcf)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)
                continue

            sols.append_line(line, sj, branch)

            # sol_key = "%3d%3d" % (line.vl, line.v2l)  # (v', v'') transition (v_sup, v_inf)
            # raise RuntimeError("Como que o calculate_qgbd esta fazendo, sendo que o dicionario molconsts agora tem prefixos to e from?")
            # if sol_key not in sols:
            #     qgbd = self._calculate_qgbd(line.v2l)
            #     sols[sol_key] = pyfant.SetOfLines(line.vl, line.v2l,
            #                                   qgbd["qv"], qgbd["gv"], qgbd["bv"], qgbd["dv"], 1.)
            #
            # sol = sols[sol_key]
            # sol.append_line(wl, gf_pfant, J2l_pfant, branch)

        return sols, log
